##########################################################
### R script to compute DMPs between histologies       ###
### (Sup Data 16) and clusters (Sup Data 11 and 17)    ###
##########################################################


## load libraries
library(GenomeInfoDb)
library(foreach)
library(stringi) 
library(XML)
library(bumphunter)
library(mclust)
library(survival) 
library(minfi)
library(ENmix)
library(FDb.InfiniumMethylation.hg19)
library(methylumi)
library(RCurl)
library(wateRmelon)
library(limma)
library(qqman)
library(DMRcatedata)
library(DMRcate)
library(TxDb.Hsapiens.UCSC.hg19.knownGene)
library(IlluminaHumanMethylationEPICanno.ilm10b2.hg19)
library(NMF)

## loading data
load("methylation.RData") #Load array data; this loads a funnometa object with pre-processed methylation data and metadata (pData). Note: methylation data are located in the European Genome-Phenome archive (EGAS#1096)
annEPIC = getAnnotation(IlluminaHumanMethylationEPICanno.ilm10b2.hg19) # load annotation

# get subset of data with both age and sex information
pData = pData(funnometa)
funnometa = funnometa[,!(is.na(pData$validatedsex)|is.na(pData$Age_class))]

## DMPs between histological types (Supplementary Data 15)
# Filtering 
beta <- minfi::getBeta(funnometa)

beta_TC <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData(funnometa)$CPR == "Typical"),]))])
beta_AC <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData(funnometa)$CPR == "Atypical"),]))])
beta_LCNEC <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData(funnometa)$CPR == "LCNEC"),]))])

beta_TC <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData(funnometa)$CPR == "Typical"),]))])
beta_AC <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData(funnometa)$CPR == "Atypical"),]))])
beta_LCNEC <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData(funnometa)$CPR == "LCNEC"),]))])

DIFF_TCAC    <- which(abs(beta_TC - beta_AC) > 0.2)
DIFF_TCLCNEC <- which(abs(beta_TC - beta_LCNEC) > 0.2)
DIFF_ACLCNEC <- which(abs(beta_AC - beta_LCNEC) > 0.2)
#clean to reduce memory usage
rm(beta_TC)
rm(beta_AC)
rm(beta_LCNEC)
gc()

meth <- minfi::getM(funnometa)
pData = pData(funnometa)
methTCAC <- meth[DIFF_TCAC,]
methTCLCNEC <- meth[DIFF_TCLCNEC,]
methACLCNEC <- meth[DIFF_ACLCNEC,]

#load sva results to correct for batch effects
load("sva90.RData")
sva <- sva[rownames(pData),] 
pData <- cbind(pData, sva)

#build model contrasts
mod <- model.matrix(~ 0 + ., data=pData[,c(6,11,31,31+1:10)])
contrast.matrix.TCAC   <-matrix(c(-1,0,1,0,rep(0,12)))
contrast.matrix.TCLCNEC<-matrix(c(0,-1,1,0,rep(0,12)))
contrast.matrix.ACLCNEC<-matrix(c(1,-1,0,0,rep(0,12)))

#fit model
fitTCAC = lmFit(methTCAC,mod)
fitContrastsTCAC = contrasts.fit(fitTCAC,contrast.matrix.TCAC)
ebTCAC = eBayes(fitContrastsTCAC)
topTCAC <- topTable(ebTCAC, adjust="BH",p.value=0.05, n=100000000)

fitTCLCNEC = lmFit(methTCLCNEC,mod)
fitContrastsTCLCNEC = contrasts.fit(fitTCLCNEC,contrast.matrix.TCLCNEC)
ebTCLCNEC = eBayes(fitContrastsTCLCNEC)
topTCLCNEC <- topTable(ebTCLCNEC, adjust="BH",p.value=0.05, n=100000000)

fitACLCNEC = lmFit(methACLCNEC,mod)
fitContrastsACLCNEC = contrasts.fit(fitACLCNEC,contrast.matrix.ACLCNEC)
ebACLCNEC = eBayes(fitContrastsACLCNEC)
topACLCNEC <- topTable(ebACLCNEC, adjust="BH",p.value=0.05, n=100000000)

## compute core DMPs
require(tidyverse)
DMPS.Histo = full_join(full_join(data.frame(cpg=rownames(topTCAC), topTCAC,stringsAsFactors = F),
                                 data.frame(cpg=rownames(topTCLCNEC),topTCLCNEC,stringsAsFactors = F),
                                 "cpg", suffix = c(".TC_vs_AC",".TC_vs_LCNEC") ), 
                       data.frame(cpg=rownames(topACLCNEC), topACLCNEC,stringsAsFactors = F),
                       suffix = c("",".AC_vs_LCNEC") )

colnames(DMPS.Histo)[14:19] = paste0(colnames(DMPS.Histo)[14:19],".AC_vs_LCNEC")

write.csv(data.frame(DMPS.Histo,core.CpG.TC=(!is.na(DMPS.Histo$adj.P.Val.TC_vs_AC)) & (!is.na(DMPS.Histo$adj.P.Val.TC_vs_LCNEC)),
                     core.CpG.AC=(!is.na(DMPS.Histo$adj.P.Val.TC_vs_AC)) & (!is.na(DMPS.Histo$adj.P.Val.AC_vs_LCNEC)),
                     core.CpG.LCNEC=(!is.na(DMPS.Histo$adj.P.Val.TC_vs_LCNEC)) & (!is.na(DMPS.Histo$adj.P.Val.AC_vs_LCNEC)) ),
          file = "TableS_DMP_Histology.csv",quote = F)

coreDMPs = DMPS.Histo

# annotate
sig <- cbind(coreDMPs, annEPIC[match(coreDMPs$cpg,rownames(annEPIC)), ])
sig$start <- sig$pos
sig$end <- sig$pos
sig.gr <- as(sig, "GenomicRanges")
nearest2<-getNearestTSS(sig.gr)
tab<-cbind(sig,nearest2)
colnames(tab)[75]<-"nearestTSS"
write.csv(tab[,-c(26:39+1,42:68+1,72,76)],file = "TableS_DMP_Histology_annot.csv",quote = F)


## DMPs between LNET clusters (Supplementary Data 11)
# Filtering 
beta <- minfi::getBeta(funnometa)
pData = pData(funnometa)

beta_A1 <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData$cluster_LNET == "Ca A1"),]))])
beta_A2 <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData$cluster_LNET == "Ca A2"),]))])
beta_B <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData$cluster_LNET == "Ca B"),]))])
DIFF_A1A2<- which(abs(beta_A1 - beta_A2) > 0.2)
DIFF_A1B <- which(abs(beta_A1 - beta_B) > 0.2)
DIFF_A2B <- which(abs(beta_A2 - beta_B) > 0.2)

#clean to reduce memory usage
rm(beta_A1)
rm(beta_A2)
rm(beta_B)
gc()

meth <- minfi::getM(funnometa)
meth = meth[,!is.na(pData$cluster_LNET)]
pData = pData(funnometa)[!is.na(pData$cluster_LNET),]

methA1A2<- meth[DIFF_A1A2,]
methA1B <- meth[DIFF_A1B,]
methA2B <- meth[DIFF_A2B,]

mod <- model.matrix(~ 0 + ., data=cbind( pData$cluster_LNET, pData[,c(11,31,31+1:10)])  )
contrast.matrix.A1A2<-matrix(c(1,-1,0,0,rep(0,12)))
contrast.matrix.A1B <-matrix(c(1,0,-1,0,rep(0,12)))
contrast.matrix.A2B <-matrix(c(0,1,-1,0,rep(0,12)))

fitA1A2 = lmFit(methA1A2,mod)
fitContrastsA1A2 = contrasts.fit(fitA1A2,contrast.matrix.A1A2)
ebA1A2 = eBayes(fitContrastsA1A2)
topA1A2 <- topTable(ebA1A2, adjust="BH",p.value=0.05, n=100000000)

fitA1B  = lmFit(methA1B,mod)
fitContrastsA1B = contrasts.fit(fitA1B,contrast.matrix.A1B)
ebA1B   = eBayes(fitContrastsA1B)
topA1B <- topTable(ebA1B, adjust="BH",p.value=0.05, n=100000000)

fitA2B = lmFit(methA2B,mod)
fitContrastsA2B = contrasts.fit(fitA2B,contrast.matrix.A2B)
ebA2B = eBayes(fitContrastsA2B)
topA2B <- topTable(ebA2B, adjust="BH",p.value=0.05, n=100000000)

# core DMPs
require(tidyverse)
DMPS.ClustLNET = full_join(full_join(data.frame(cpg=rownames(topA1A2), topA1A2,stringsAsFactors = F), data.frame(cpg=rownames(topA1B),topA1B,stringsAsFactors = F),
                                     "cpg", suffix = c(".A1_vs_A2",".A1_vs_B") ), data.frame(cpg=rownames(topA2B), topA2B,stringsAsFactors = F), suffix = c("",".A2_vs_B") )

colnames(DMPS.ClustLNET)[14:19] = paste0(colnames(DMPS.ClustLNET)[14:19],".A2_vs_B")

write.csv(data.frame(DMPS.ClustLNET,core.CpG.A1=(!is.na(DMPS.ClustLNET$adj.P.Val.A1_vs_A2)) & (!is.na(DMPS.ClustLNET$adj.P.Val.A1_vs_B)),
                     core.CpG.A2=(!is.na(DMPS.ClustLNET$adj.P.Val.A1_vs_A2)) & (!is.na(DMPS.ClustLNET$adj.P.Val.A2_vs_B)),
                     core.CpG.B=(!is.na(DMPS.ClustLNET$adj.P.Val.A1_vs_B)) & (!is.na(DMPS.ClustLNET$adj.P.Val.A2_vs_B)) ),
          file = "TableS_DMP_ClustLNET.csv",quote = F,row.names = F)

coreDMPs = DMPS.ClustLNET

# annotate
sig <- cbind(coreDMPs, annEPIC[match(coreDMPs$cpg,rownames(annEPIC)), ])
sig$start <- sig$pos
sig$end <- sig$pos
sig.gr <- as(sig, "GenomicRanges")
nearest2<-getNearestTSS(sig.gr)
tab<-cbind(sig,nearest2)
head(tab)
colnames(tab)[75]<-"nearestTSS"
write.csv(tab[,-c(26:39+1,42:68+1,72,76)],file = "TableS_DMP_Histology_annot.csv",quote = F)


## DMPs between LNEN clusters (Supplementary Data 16)
# Filtering 
beta <- minfi::getBeta(funnometa)
pData = pData(funnometa)
beta_A <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData$cluster_LNEN == "Ca A"),]))])
beta_B <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData$cluster_LNEN == "Ca B"),]))])
beta_LCNEC <- rowMeans(beta[,which(colnames(beta) %in% rownames(pData(funnometa)[which(pData$cluster_LNEN == "LCNEC"),]))])
DIFF_AB    <- which(abs(beta_A - beta_B) > 0.2)
DIFF_ALCNEC <- which(abs(beta_A - beta_LCNEC) > 0.2)
DIFF_BLCNEC <- which(abs(beta_B - beta_LCNEC) > 0.2)

rm(beta)
rm(beta_A)
rm(beta_B)
rm(beta_LCNEC)
gc()

meth <- minfi::getM(funnometa)
rm(funnometa)

methAB <- meth[DIFF_AB,]
methALCNEC <- meth[DIFF_ALCNEC,]
methBLCNEC <- meth[DIFF_BLCNEC,]
rm(meth)
gc()

pData <- cbind(pData, sva)
mod <- model.matrix(~ 0 + ., data=cbind( pData$cluster_LNEN, pData[,c(11,31,31+1:10)])  )

contrast.matrix.AB   <-matrix(c(1,-1,0,0,rep(0,12)))
contrast.matrix.ALCNEC<-matrix(c(1,0,-1,0,rep(0,12)))
contrast.matrix.BLCNEC<-matrix(c(0,1,-1,0,rep(0,12)))

fitAB  = lmFit(methAB,mod)
fitContrastsAB = contrasts.fit(fitAB,contrast.matrix.AB)
ebAB   = eBayes(fitContrastsAB)
topAB <- topTable(ebAB, adjust="BH",p.value=0.05, n=100000000)

fitALCNEC = lmFit(methALCNEC,mod)
fitContrastsALCNEC = contrasts.fit(fitALCNEC,contrast.matrix.ALCNEC)
ebALCNEC = eBayes(fitContrastsALCNEC)
topALCNEC <- topTable(ebALCNEC, adjust="BH",p.value=0.05, n=100000000)

fitBLCNEC = lmFit(methBLCNEC,mod)
fitContrastsBLCNEC = contrasts.fit(fitBLCNEC,contrast.matrix.BLCNEC)
ebBLCNEC = eBayes(fitContrastsBLCNEC)
topBLCNEC <- topTable(ebBLCNEC, adjust="BH",p.value=0.05, n=100000000)

## core DMPs
require(tidyverse)
DMPS.ClustLNEN = full_join(full_join(data.frame(cpg=rownames(topAB), topAB,stringsAsFactors = F), data.frame(cpg=rownames(topALCNEC),topALCNEC,stringsAsFactors = F),
                                 "cpg", suffix = c(".A_vs_B",".A_vs_LCNEC") ), data.frame(cpg=rownames(topBLCNEC), topBLCNEC,stringsAsFactors = F), suffix = c("",".B_vs_LCNEC") )

colnames(DMPS.ClustLNEN)[14:19] = paste0(colnames(DMPS.ClustLNEN)[14:19],".B_vs_LCNEC")

write.csv(data.frame(DMPS.ClustLNEN,core.CpG.A=(!is.na(DMPS.ClustLNEN$adj.P.Val.A_vs_B)) & (!is.na(DMPS.ClustLNEN$adj.P.Val.A_vs_B)),
                     core.CpG.B=(!is.na(DMPS.ClustLNEN$adj.P.Val.A_vs_B)) & (!is.na(DMPS.ClustLNEN$adj.P.Val.A_vs_LCNEC)),
                     core.CpG.LCNEC=(!is.na(DMPS.ClustLNEN$adj.P.Val.A_vs_LCNEC)) & (!is.na(DMPS.ClustLNEN$adj.P.Val.B_vs_LCNEC)) ),
          file = "TableS_DMP_ClustLNEN.csv",quote = F,row.names = F)


coreDMPs = read.csv("TableS_DMP_ClustLNEN.csv")

# annotate
sig <- cbind(coreDMPs, annEPIC[match(coreDMPs$cpg,rownames(annEPIC)), ])
sig$start <- sig$pos
sig$end <- sig$pos
sig.gr <- as(sig, "GenomicRanges")
nearest2<-getNearestTSS(sig.gr)
tab<-cbind(sig,nearest2)
head(tab)
colnames(tab)[75]<-"nearestTSS"
write.csv(tab[,-c(26:39,42:68,71,75)],file = "TableS_DMP_ClustLNEN_annot.csv",quote = F)


