##################################################################
### Classification functions used to run Machine Learning      ###
### analyses from Fig 1, Sup Figs 9-12 and 31, and Sup Data 1  ###
##################################################################

# function to run leave one out with variable selection and normalization on the training set 
# this function was used to perform the ML classification based on methylation features
# Args: 
#   data_to_classify: a matrix (samples in rows and features in column)
#   cancer_to_classify: vector containing the labels for each sample
#   cv: a list of indices to consider in each fold
#   class_type: type of classification indicating which tumor types are included in the analysis, in the paper we always used "LCNECvsACvsTC"
# Optional args: 
#   filter_var: amount of variance to be explained by the features kept in the model (variable selection); uc_threshold: ratio to define the UC category; 
#   nb_cores: number of cores to use for parallelization; title: title of the confusion matrix plot; n_top: number of most important features to save at each fold; 
#   resample_function: function to use for resampling; print_out: boolean, set to TRUE to plot the confusion matrix;
#   low_grad: color corresponding to the lowest values in the confusion matrix plot; high_grad: color corresponding to the highest values in the confusion matrix plot;
#   family_type: font used in the confusion matrix plot
# Outputs: 
#   confusion matrix based on the predictions
#   vector containing the 10 most important features for each fold 
#   summary of the predictions probabilities
#   if print_out=T, the confusion matrix based on the predictions will be plotted
classif_varSelect_norm=function(data_to_classify,cancer_to_classify,cv,class_type,
                                filter_var=0.05,uc_threshold=1.5,nb_cores=7,title="",n_top=10,resample_function="oversample",
                                print_out=T,low_grad="#ece2f0",high_grad="#1c9099",family_type="Times"){
  cl=makeCluster(as.numeric(nb_cores), outfile="")
  registerDoParallel(cl) # Use nb_cores cores for parallelisation

  resample=get(resample_function)
  results <- foreach(fold = cv, .packages="randomForest") %dopar% {
    
    testData=data_to_classify[fold,,drop=F]
    cancerTest=cancer_to_classify[fold]
    
    if(class_type=="LCNECvsPCA"){
      #remove duplicated sample in training 
      if("S00716_A" %in% rownames(data_to_classify)[fold]){
        trainData=data_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_B")),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_B"))]))
      }else if("S00716_B" %in% rownames(data_to_classify)[fold]){
        trainData=data_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_A" )),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_A"))]))
      }else{
        trainData=data_to_classify[-c(fold),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold)]))
      }
    }else{
      #remove duplicated sample and carcinoids in training 
      if("S00716_A" %in% rownames(data_to_classify)[fold]){
        trainData=data_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_B" | cancer_to_classify=="Carcinoid")),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"),which(rownames(data_to_classify)=="S00716_B"))]))
      }else if("S00716_B" %in% rownames(data_to_classify)[fold]){
        trainData=data_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_A" | cancer_to_classify=="Carcinoid")),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"),which(rownames(data_to_classify)=="S00716_A"))]))
      }else{
        trainData=data_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid")),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"))]))
      }
    }
    
    # variable selection
    vR  = apply(trainData,2,var)
    cR  = cumsum( vR[rev(order(vR)) ] )/sum(vR)
    idR = rev(which( cR<=filter_var))[1]
    trainfiltered  = trainData[,rev(order(vR))[1:idR]]
    testfiltered   = testData[,rev(order(vR))[1:idR],drop=F]
    
    # normalization
    data_means = colMeans( trainfiltered )
    data_sd    = apply( trainfiltered,2,sd )
    trainData=scale(trainfiltered,center = T,scale = T)
    testData=scale(testfiltered,center = data_means,scale = data_sd)
    
    resampling=resample(trainData,cancerTrain)
    trainData=resampling[[1]]
    cancerTrain=resampling[[2]]
    
    model=randomForest(trainData,cancerTrain,importance=T,do.trace=F,probs=T) 
    top=rownames(model$importance[order(model$importance[,(length(levels(cancerTrain))+1)],decreasing = T),,drop=F])[1:n_top]
    list(accuracy=data.frame(pred=predict(model,testData),cohort=cancerTest),proba=predict(model,testData,type="prob"),top_genes=top)
  }

  stopCluster(cl)
  
  # Merge folds results
  merge_results=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$accuracy) 
  }
  probabilities=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$proba) 
  }
  top=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$top) 
  }
  
  # Compute confusion matrix
  res=cbind(probabilities,merge_results)
  res$pred=as.vector(res$pred)
  res$cohort=as.vector(res$cohort)
  res$class=(res$pred==res$cohort)
  # define the unclassified category
  res$maxs_diff=sapply(1:nrow(res), function(i) res[i,order(res[i,1:length(unique(res$pred))],decreasing = T)[1]]/res[i,order(res[i,1:length(unique(res$pred))],decreasing = T)[2]] >=uc_threshold)
  levels(merge_results$pred) <- c(levels(merge_results$pred),"UC")
  merge_results$pred[which(res$maxs_diff==FALSE)] <- "UC"
  #remove duplicates and carcinoids for the confusion_matrix plot
  to_remove=which(merge_results$cohort=="Carcinoid" | rownames(merge_results) == "S00716_A")
  if(length(to_remove)!=0){
    merge_results=merge_results[-which(merge_results$cohort=="Carcinoid" | rownames(merge_results) == "S00716_A"),]
  }
  merge_results$cohort=as.vector(merge_results$cohort) #to avoid removing UC column in the confusion matrix when there are no UC
  confusion_matrix=t(table(merge_results$pred,merge_results$cohort))
  
  if(print_out==T){
    plot_confusion_matrix_ggplot(confusion_matrix,as.factor(merge_results$cohort),low_grad,high_grad,title,family_type,classification = class_type)
  }
  
  return(list(confusion_matrix,top,res))
}


# function to run leave one out on expression data with vst computation, variable selection and normalization on the training set
# this function was used to perform the ML classification based on expression features
# Args: 
#   data_to_classify: a DESeq object
#   cancer_to_classify: vector containing the labels for each sample
#   cv: a list of indices to consider in each fold
#   class_type: type of classification indicating which tumor types are included in the analysis, in the paper we always used "LCNECvsACvsTC"
# Optional args: the optional arguments are in common with the ones in the previous function
# Outputs: 
#   confusion matrix based on the predictions
#   vector containing the 10 most important features for each fold 
#   summary of the predictions probabilities
#   if print_out=T, the confusion matrix based on the predictions will be plotted
classif_expression_vst=function(data_to_classify,cancer_to_classify,cv,class_type,
                                filter_var=0.5,uc_threshold=1.5,nb_cores=7,title="",n_top=10,resample_function="oversample",
                                print_out=T,low_grad="#ece2f0",high_grad="#1c9099",family_type="Times"){
  cl=makeCluster(as.numeric(nb_cores), outfile="")
  registerDoParallel(cl) # Use nb_cores cores for parallelisation
  
  resample=get(resample_function)
  results <- foreach(fold = cv, .packages=c("randomForest","DESeq2")) %dopar% {
    
    testData=data_to_classify[,fold,drop=F]
    cancerTest=cancer_to_classify[fold]
    
    if(class_type=="LCNECvsPCA"){
      #remove duplicated sample in training 
      if("S00716_A" %in% colnames(data_to_classify)[fold]){
        trainData=data_to_classify[,-c(fold,which(rownames(data_to_classify)=="S00716_B")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_B"))]))
      }else if("S00716_B" %in% colnames(data_to_classify)[fold]){
        trainData=data_to_classify[,-c(fold,which(rownames(data_to_classify)=="S00716_A" )),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_A"))]))
      }else{
        trainData=data_to_classify[,-c(fold),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold)]))
      }
    }else{
      #remove duplicated sample in training 
      if("S00716_A" %in% colnames(data_to_classify)[fold]){
        trainData=data_to_classify[,-c(fold,which(colnames(data_to_classify)=="S00716_B" | cancer_to_classify=="Carcinoid")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"),which(colnames(data_to_classify)=="S00716_B"))]))
      }else if("S00716_B" %in% colnames(data_to_classify)[fold]){
        trainData=data_to_classify[,-c(fold,which(colnames(data_to_classify)=="S00716_A" | cancer_to_classify=="Carcinoid")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"),which(colnames(data_to_classify)=="S00716_A"))]))
      }else{
        trainData=data_to_classify[,-c(fold,which(cancer_to_classify=="Carcinoid")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"))]))
      }
    }

    # use trainData to retrieve size factor of testData
    trainData <- estimateSizeFactors(trainData)
    # apply sizefactors learned from training to test
    counts.train = counts(trainData)
    counts.test  = counts(testData)
    loggeomeans.train <- rowMeans(log(counts.train))
    locfunc = stats::median
    sf <- apply(counts.test, 2, function(cnts) {
      exp(locfunc((log(cnts) - loggeomeans.train)[is.finite(loggeomeans.train) & cnts > 0]))
    })
    testData@colData$sizeFactor = sf
    
    # train VST
    design(trainData) <- ~1
    design(testData) <- ~1
    trainData <- estimateDispersionsGeneEst(trainData, quiet = TRUE)
    trainData <- estimateDispersionsFit(trainData, quiet = TRUE, fitType = "parametric")
    vsd.train <- getVarianceStabilizedData(trainData)
    
    dispersionFunction(testData) <- dispersionFunction(trainData)
    vsd.test <- assay(varianceStabilizingTransformation(testData, blind=FALSE) )
    sample=colnames(vsd.test)
    
    vR= apply(vsd.train,1,var)
    cR = cumsum( vR[rev(order(vR)) ] )/sum(vR)
    idR  = rev(which( cR<=filter_var))[1]
    vsd.trainVST50  = vsd.train[rev(order(vR))[1:idR],]
    vsd.testVST50=vsd.test[rev(order(vR))[1:idR],]

    # normalization
    vst_means=rowMeans( vsd.trainVST50 ) 
    vst_sd=apply( vsd.trainVST50,1,sd ) 
    trainData=scale(t(vsd.trainVST50),center = T,scale = T)
    testData=scale(t(vsd.testVST50),center = vst_means,scale = vst_sd)
    rownames(testData)=sample
    
    resampling=resample(trainData,cancerTrain)
    trainData=resampling[[1]]
    cancerTrain=resampling[[2]]
    
    model=randomForest(trainData,cancerTrain,importance=T,do.trace=F,probs=T) 
    top=rownames(model$importance[order(model$importance[,(length(levels(cancerTrain))+1)],decreasing = T),,drop=F])[1:n_top]
    list(accuracy=data.frame(pred=predict(model,testData),cohort=cancerTest),proba=predict(model,testData,type="prob"),top_genes=top)
  }

  stopCluster(cl)
  
  # Merge folds results
  merge_results=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$accuracy) 
  }
  probabilities=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$proba) 
  }
  top=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$top) 
  }
  
  # Compute confusion matrix
  res=cbind(probabilities,merge_results)
  res$pred=as.vector(res$pred)
  res$cohort=as.vector(res$cohort)
  res$class=(res$pred==res$cohort)
  res$maxs_diff=sapply(1:nrow(res), function(i) res[i,order(res[i,1:length(unique(res$pred))],decreasing = T)[1]]/res[i,order(res[i,1:length(unique(res$pred))],decreasing = T)[2]] >=uc_threshold)
  
  levels(merge_results$pred) <- c(levels(merge_results$pred),"UC")
  merge_results$pred[which(res$maxs_diff==FALSE)] <- "UC"
  #remove duplicates and carcinoids for the confusion_matrix plot
  to_remove=which(merge_results$cohort=="Carcinoid" | rownames(merge_results) == "S00716_A")
  if(length(to_remove)!=0){
    merge_results=merge_results[-which(merge_results$cohort=="Carcinoid" | rownames(merge_results) == "S00716_A"),]
  }
  merge_results$cohort=as.vector(merge_results$cohort) #to avoid removing UC column in the confusion matrix when there are no UC
  confusion_matrix=t(table(merge_results$pred,merge_results$cohort))
  
  if(print_out==T){
    plot_confusion_matrix_ggplot(confusion_matrix,as.factor(merge_results$cohort),low_grad,high_grad,title,family_type,classification = class_type)
  }
  
  return(list(confusion_matrix,top,res))
}

# function to run leave one out on MKI67 expression data with vst computation, variable selection and normalization on the training set
# Args: 
#   data_to_classify: a DESeq object
#   cancer_to_classify: vector containing the labels for each sample
#   cv: a list of indices to consider in each fold
#   class_type: type of classification indicating which tumor types are included in the analysis, in the paper we always used "LCNECvsACvsTC"
# Optional args: the optional arguments are in common with the ones in the previous function
# Outputs: 
#   confusion matrix based on the predictions
#   vector containing the 10 most important features for each fold 
#   summary of the predictions probabilities
#   if print_out=T, the confusion matrix based on the predictions will be plotted
classif_MKI67_vst=function(data_to_classify,cancer_to_classify,cv,class_type,
                           uc_threshold=1.5,nb_cores=7,title="",n_top=10,resample_function="oversample",
                           print_out=T,low_grad="#ece2f0",high_grad="#1c9099",family_type="Times"){
  cl=makeCluster(as.numeric(nb_cores), outfile="")
  registerDoParallel(cl) # Use nb_cores cores for parallelisation

  resample=get(resample_function)
  results <- foreach(fold = cv, .packages=c("randomForest","DESeq2")) %dopar% {
    
    testData=data_to_classify[,fold,drop=F]
    cancerTest=cancer_to_classify[fold]
    
    if(class_type=="LCNECvsPCA"){
      #remove duplicated sample in training 
      if("S00716_A" %in% colnames(data_to_classify)[fold]){
        trainData=data_to_classify[,-c(fold,which(rownames(data_to_classify)=="S00716_B")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_B"))]))
      }else if("S00716_B" %in% colnames(data_to_classify)[fold]){
        trainData=data_to_classify[,-c(fold,which(rownames(data_to_classify)=="S00716_A" )),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_A"))]))
      }else{
        trainData=data_to_classify[,-c(fold),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold)]))
      }
    }else{
      #remove duplicated sample in training 
      if("S00716_A" %in% colnames(data_to_classify)[fold]){
        trainData=data_to_classify[,-c(fold,which(colnames(data_to_classify)=="S00716_B" | cancer_to_classify=="Carcinoid")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"),which(colnames(data_to_classify)=="S00716_B"))]))
      }else if("S00716_B" %in% colnames(data_to_classify)[fold]){
        trainData=data_to_classify[,-c(fold,which(colnames(data_to_classify)=="S00716_A" | cancer_to_classify=="Carcinoid")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"),which(colnames(data_to_classify)=="S00716_A"))]))
      }else{
        trainData=data_to_classify[,-c(fold,which(cancer_to_classify=="Carcinoid")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"))]))
      }
    }

    # use trainData to retrieve size factor of testData
    trainData <- estimateSizeFactors(trainData)
    # apply sizefactors learned from training to test
    counts.train = counts(trainData)
    counts.test  = counts(testData)
    loggeomeans.train <- rowMeans(log(counts.train))
    locfunc = stats::median
    sf <- apply(counts.test, 2, function(cnts) {
      exp(locfunc((log(cnts) - loggeomeans.train)[is.finite(loggeomeans.train) & cnts > 0]))
    })
    testData@colData$sizeFactor = sf
    
    # train VST
    design(trainData) <- ~1
    design(testData) <- ~1
    trainData <- estimateDispersionsGeneEst(trainData, quiet = TRUE)
    trainData <- estimateDispersionsFit(trainData, quiet = TRUE, fitType = "parametric")
    vsd.train <- getVarianceStabilizedData(trainData)
    
    dispersionFunction(testData) <- dispersionFunction(trainData)
    vsd.test <- assay(varianceStabilizingTransformation(testData, blind=FALSE) )

    # keep only MKI67 expression data
    trainData  = t(vsd.train["MKI67",,drop=F])
    testData   = t(vsd.test["MKI67",,drop=F])

    resampling=resample(trainData,cancerTrain)
    trainData=resampling[[1]]
    cancerTrain=resampling[[2]]
    
    model=randomForest(trainData,cancerTrain,importance=T,do.trace=F,probs=T) 
    top=rownames(model$importance[order(model$importance[,(length(levels(cancerTrain))+1)],decreasing = T),,drop=F])[1:n_top]
    list(accuracy=data.frame(pred=predict(model,testData),cohort=cancerTest),proba=predict(model,testData,type="prob"),top_genes=top)
  }
  stopCluster(cl)
  
  # Merge folds results
  merge_results=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$accuracy) 
  }
  probabilities=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$proba) 
  }
  top=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$top) 
  }
  
  # Compute confusion matrix
  res=cbind(probabilities,merge_results)
  res$pred=as.vector(res$pred)
  res$cohort=as.vector(res$cohort)
  res$class=(res$pred==res$cohort)
  res$maxs_diff=sapply(1:nrow(res), function(i) res[i,order(res[i,1:length(unique(res$pred))],decreasing = T)[1]]/res[i,order(res[i,1:length(unique(res$pred))],decreasing = T)[2]] >=uc_threshold)
  
  levels(merge_results$pred) <- c(levels(merge_results$pred),"UC")
  merge_results$pred[which(res$maxs_diff==FALSE)] <- "UC"
  #remove duplicates and carcinoids for the confusion_matrix plot
  to_remove=which(merge_results$cohort=="Carcinoid" | rownames(merge_results) == "S00716_A")
  if(length(to_remove)!=0){
    merge_results=merge_results[-which(merge_results$cohort=="Carcinoid" | rownames(merge_results) == "S00716_A"),]
  }
  merge_results$cohort=as.vector(merge_results$cohort) #to avoid removing UC column in the confusion matrix when there are no UC
  confusion_matrix=t(table(merge_results$pred,merge_results$cohort))
  
  if(print_out==T){
    plot_confusion_matrix_ggplot(confusion_matrix,as.factor(merge_results$cohort),low_grad,high_grad,title,family_type,classification = class_type)
  }
  
  return(list(confusion_matrix,top,res))
}

# function to run leave one out on expression and methylation data with vst computation, variable selection and normalization on the training set
# Args: 
#   data_to_classify_expression: a DESeq object
#   data_to_classify_methylation: a matrix (samples in rows and features in column) containing the M values
#   cancer_to_classify: vector containing the labels for each sample
#   cv: a list of indices to consider in each fold
#   class_type: type of classification indicating which tumor types are included in the analysis, in the paper we always used "LCNECvsACvsTC"
# Optional args: the optional arguments are in common with the ones in the previous function
# Outputs: 
#   confusion matrix based on the predictions
#   vector containing the 10 most important features for each fold 
#   summary of the predictions probabilities
#   if print_out=T, the confusion matrix based on the predictions will be plotted
classif_expression_methylation=function(data_to_classify_expression,data_to_classify_methylation,cancer_to_classify,cv,class_type,
                                uc_threshold=1.5,nb_cores=7,title="",n_top=10,resample_function="oversample",
                                print_out=T,low_grad="#ece2f0",high_grad="#1c9099",family_type="Times"){
  cl=makeCluster(as.numeric(nb_cores), outfile="")
  registerDoParallel(cl) # Use nb_cores cores for parallelisation
  
  resample=get(resample_function)
  results <- foreach(fold = cv, .packages=c("randomForest","DESeq2")) %dopar% {
    
    #expression data
    testData_exp=data_to_classify_expression[,fold,drop=F]
    cancerTest=cancer_to_classify[fold]
    
    if(class_type=="LCNECvsPCA"){
      #remove duplicated sample in training 
      if("S00716_A" %in% colnames(data_to_classify_expression)[fold]){
        trainData_exp=data_to_classify_expression[,-c(fold,which(rownames(data_to_classify_expression)=="S00716_B")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(rownames(data_to_classify_expression)=="S00716_B"))]))
      }else if("S00716_B" %in% colnames(data_to_classify_expression)[fold]){
        trainData_exp=data_to_classify_expression[,-c(fold,which(rownames(data_to_classify_expression)=="S00716_A" )),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(rownames(data_to_classify_expression)=="S00716_A"))]))
      }else{
        trainData_exp=data_to_classify_expression[,-c(fold),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold)]))
      }
    }else{
      #remove duplicated sample in training 
      if("S00716_A" %in% colnames(data_to_classify_expression)[fold]){
        trainData_exp=data_to_classify_expression[,-c(fold,which(colnames(data_to_classify_expression)=="S00716_B" | cancer_to_classify=="Carcinoid")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"),which(colnames(data_to_classify_expression)=="S00716_B"))]))
      }else if("S00716_B" %in% colnames(data_to_classify_expression)[fold]){
        trainData_exp=data_to_classify_expression[,-c(fold,which(colnames(data_to_classify_expression)=="S00716_A" | cancer_to_classify=="Carcinoid")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"),which(colnames(data_to_classify_expression)=="S00716_A"))]))
      }else{
        trainData_exp=data_to_classify_expression[,-c(fold,which(cancer_to_classify=="Carcinoid")),drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"))]))
      }
    }
    
    # use trainData to retrieve size factor of testData
    trainData_exp <- estimateSizeFactors(trainData_exp)
    # apply sizefactors learned from training to test
    counts.train = counts(trainData_exp)
    counts.test  = counts(testData_exp)
    loggeomeans.train <- rowMeans(log(counts.train))
    locfunc = stats::median
    sf <- apply(counts.test, 2, function(cnts) {
      exp(locfunc((log(cnts) - loggeomeans.train)[is.finite(loggeomeans.train) & cnts > 0]))
    })
    testData_exp@colData$sizeFactor = sf
    
    # train VST
    design(trainData_exp) <- ~1
    design(testData_exp) <- ~1
    trainData_exp <- estimateDispersionsGeneEst(trainData_exp, quiet = TRUE)
    trainData_exp <- estimateDispersionsFit(trainData_exp, quiet = TRUE, fitType = "parametric")
    vsd.train <- getVarianceStabilizedData(trainData_exp)
    
    dispersionFunction(testData_exp) <- dispersionFunction(trainData_exp)
    vsd.test <- assay(varianceStabilizingTransformation(testData_exp, blind=FALSE) )
    sample=colnames(vsd.test)
    
    vR= apply(vsd.train,1,var)
    cR = cumsum( vR[rev(order(vR)) ] )/sum(vR)
    idR  = rev(which( cR<=0.5))[1]
    vsd.trainVST50  = vsd.train[rev(order(vR))[1:idR],]
    vsd.testVST50=vsd.test[rev(order(vR))[1:idR],]
    
    # methylation
    testData_methyl <- data_to_classify_methylation[fold,,drop=F]

    if(class_type=="LCNECvsPCA"){
      #remove duplicated sample in training 
      if("S00716_A" %in% rownames(data_to_classify_methylation)[fold]){
        trainData_methyl=data_to_classify_methylation[-c(fold,which(rownames(data_to_classify_methylation)=="S00716_B")),,drop=F]
      }else if("S00716_B" %in% rownames(data_to_classify_methylation)[fold]){
        trainData_methyl=data_to_classify_methylation[-c(fold,which(rownames(data_to_classify_methylation)=="S00716_A" )),,drop=F]
      }else{
        trainData_methyl=data_to_classify_methylation[-c(fold),,drop=F]
      }
    }else{
      #remove duplicated sample and carcinoids in training 
      if("S00716_A" %in% rownames(data_to_classify_methylation)[fold]){
        trainData_methyl=data_to_classify_methylation[-c(fold,which(rownames(data_to_classify_methylation)=="S00716_B" | cancer_to_classify=="Carcinoid")),,drop=F]
      }else if("S00716_B" %in% rownames(data_to_classify_methylation)[fold]){
        trainData_methyl=data_to_classify_methylation[-c(fold,which(rownames(data_to_classify_methylation)=="S00716_A" | cancer_to_classify=="Carcinoid")),,drop=F]
      }else{
        trainData_methyl=data_to_classify_methylation[-c(fold,which(cancer_to_classify=="Carcinoid")),,drop=F]
      }
    }
    
    # variable selection
    vR  = apply(trainData_methyl,2,var)
    cR  = cumsum( vR[rev(order(vR)) ] )/sum(vR)
    idR = rev(which( cR<=0.05))[1]
    trainfiltered_methyl  = trainData_methyl[,rev(order(vR))[1:idR]]
    testfiltered_methyl   = testData_methyl[,rev(order(vR))[1:idR],drop=F]
    
    # merge both datasets
    testData_merged=t((c(vsd.testVST50,testfiltered_methyl)))
    trainData_merged=cbind(t(vsd.trainVST50),trainfiltered_methyl)
    
    # normalization
    data_means = colMeans( trainData_merged )
    data_sd    = apply( trainData_merged,2,sd )
    trainData=scale(trainData_merged,center = T,scale = T)
    testData=scale(testData_merged,center = data_means,scale = data_sd)
    
    rownames(testData)=sample
    colnames(testData)=colnames(trainData)
    resampling=resample(trainData,cancerTrain)
    trainData=resampling[[1]]
    cancerTrain=resampling[[2]]
    
    model=randomForest(trainData,cancerTrain,importance=T,do.trace=F,probs=T) 
    top=rownames(model$importance[order(model$importance[,(length(levels(cancerTrain))+1)],decreasing = T),,drop=F])[1:n_top]
    list(accuracy=data.frame(pred=predict(model,testData),cohort=cancerTest),proba=predict(model,testData,type="prob"),top_genes=top)
  }
  
  stopCluster(cl)
  
  # Merge folds results
  merge_results=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$accuracy) 
  }
  probabilities=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$proba) 
  }
  top=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$top) 
  }
  
  # Compute confusion matrix
  res=cbind(probabilities,merge_results)
  res$pred=as.vector(res$pred)
  res$cohort=as.vector(res$cohort)
  res$class=(res$pred==res$cohort)
  res$maxs_diff=sapply(1:nrow(res), function(i) res[i,order(res[i,1:length(unique(res$pred))],decreasing = T)[1]]/res[i,order(res[i,1:length(unique(res$pred))],decreasing = T)[2]] >=uc_threshold)
  
  levels(merge_results$pred) <- c(levels(merge_results$pred),"UC")
  merge_results$pred[which(res$maxs_diff==FALSE)] <- "UC"
  #remove duplicates and carcinoids for the confusion_matrix plot
  to_remove=which(merge_results$cohort=="Carcinoid" | rownames(merge_results) == "S00716_A")
  if(length(to_remove)!=0){
    merge_results=merge_results[-which(merge_results$cohort=="Carcinoid" | rownames(merge_results) == "S00716_A"),]
  }
  merge_results$cohort=as.vector(merge_results$cohort) #to avoid removing UC column in the confusion matrix when there are no UC
  confusion_matrix=t(table(merge_results$pred,merge_results$cohort))
  
  if(print_out==T){
    plot_confusion_matrix_ggplot(confusion_matrix,as.factor(merge_results$cohort),low_grad,high_grad,title,family_type,classification = class_type)
  }
  
  return(list(confusion_matrix,top,res))
}

# function to run leave one out without variable selection and normalization on the training set
# this function was used to perform the ML classification based on MOFA and PCA features 
# Args: 
#   data_to_classify: a matrix (samples in rows and features in column)
#   cancer_to_classify: vector containing the labels for each sample
#   cv: a list of indices to consider in each fold
#   class_type: type of classification indicating which tumor types are included in the analysis, in the paper we always used "LCNECvsACvsTC"
# Optional args: the optional arguments are in common with the ones in the previous function
# Outputs: 
#   confusion matrix based on the predictions
#   vector containing the 10 most important features for each fold 
#   summary of the predictions probabilities
#   if print_out=T, the confusion matrix based on the predictions will be plotted
classif_NO_varSelect_norm=function(data_to_classify,cancer_to_classify,cv,class_type,
                                   uc_threshold=1.5,nb_cores=7,title="",n_top=10,resample_function="oversample",
                                   print_out=T,low_grad="#ece2f0",high_grad="#1c9099",family_type="Times"){
  cl=makeCluster(as.numeric(nb_cores), outfile="")
  registerDoParallel(cl) # Use nb_cores cores for parallelisation
  
  resample=get(resample_function)
  results <- foreach(fold = cv, .packages="randomForest") %dopar% {
    
    testData <- data_to_classify[fold,,drop=F]
    cancerTest=cancer_to_classify[fold]
    
    if(class_type=="LCNECvsPCA"){
      #remove duplicated sample in training 
      if("S00716_A" %in% rownames(data_to_classify)[fold]){
        trainData=data_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_B")),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_B"))]))
      }else if("S00716_B" %in% rownames(data_to_classify)[fold]){
        trainData=data_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_A" )),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_A"))]))
      }else{
        trainData=data_to_classify[-c(fold),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold)]))
      }
    }else{
      #remove duplicated sample in training 
      if("S00716_A" %in% rownames(data_to_classify)[fold]){
        trainData=data_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_B" | cancer_to_classify=="Carcinoid")),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"),which(rownames(data_to_classify)=="S00716_B"))]))
      }else if("S00716_B" %in% rownames(data_to_classify)[fold]){
        trainData=data_to_classify[-c(fold,which(rownames(data_to_classify)=="S00716_A" | cancer_to_classify=="Carcinoid")),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"),which(rownames(data_to_classify)=="S00716_A"))]))
      }else{
        trainData=data_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid")),,drop=F]
        cancerTrain=as.factor(as.character(cancer_to_classify[-c(fold,which(cancer_to_classify=="Carcinoid"))]))
      }
    }
    
    resampling=resample(trainData,cancerTrain)
    trainData=resampling[[1]]
    cancerTrain=resampling[[2]]
    
    model=randomForest(trainData,cancerTrain,importance=T,do.trace=F,probs=T) 
    top=rownames(model$importance[order(model$importance[,(length(levels(cancerTrain))+1)],decreasing = T),,drop=F])[1:n_top]
    list(accuracy=data.frame(pred=predict(model,testData),cohort=cancerTest),proba=predict(model,testData,type="prob"),top_genes=top)
  }
  
  stopCluster(cl)
  
  # Merge folds results
  merge_results=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$accuracy) 
  }
  probabilities=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$proba) 
  }
  top=foreach(fold.result=results, fold.num=icount(), .combine=rbind) %do%{
    as.data.frame(fold.result$top) 
  }
  
  # Compute confusion matrix
  res=cbind(probabilities,merge_results)
  res$pred=as.vector(res$pred)
  res$cohort=as.vector(res$cohort)
  res$class=(res$pred==res$cohort)
  res$maxs_diff=sapply(1:nrow(res), function(i) res[i,order(res[i,1:length(unique(res$pred))],decreasing = T)[1]]/res[i,order(res[i,1:length(unique(res$pred))],decreasing = T)[2]] >=uc_threshold)
  levels(merge_results$pred) <- c(levels(merge_results$pred),"UC")
  merge_results$pred[which(res$maxs_diff==FALSE)] <- "UC"
  #remove duplicates and carcinoids for the confusion_matrix plot
  to_remove=which(merge_results$cohort=="Carcinoid" | rownames(merge_results) == "S00716_A")
  if(length(to_remove)!=0){
    merge_results=merge_results[-which(merge_results$cohort=="Carcinoid" | rownames(merge_results) == "S00716_A"),]
  }
  merge_results$cohort=as.vector(merge_results$cohort) #to avoid removing UC column in the confusion matrix when there are no UC
  confusion_matrix=t(table(merge_results$pred,merge_results$cohort))
  
  if(print_out==T){
    plot_confusion_matrix_ggplot(confusion_matrix,as.factor(merge_results$cohort),low_grad,high_grad,title,family_type,classification = class_type)
  }
  
  return(list(confusion_matrix,top,res))
}


# function to oversample the training set, it will balance the classes based on the sample size of the biggest group
# Args: 
#   data: a matrix (samples in rows and features in column) containing the training samples 
#   cancer: vector containing the labels for each sample in the trainsing set
# Outputs: 
#   new trainsing set after resampling
#   vector containing the labels for each sample in the oversampled trainsing set 
oversample=function(data,cancer){
  max_nb=max(summary(cancer))
  cancer_sampling=vector()
  new_data=vector()
  for(c in levels(cancer)){
    if(summary(cancer)[c] == max_nb){
      new_data=rbind(new_data,data[which(cancer==c),,drop=F])
      cancer_sampling=c(as.character(cancer_sampling),as.character(cancer[which(cancer==c)]))
    }else{
      id=c(which(cancer==c),sample(which(cancer==c),max_nb-length(which(cancer==c)),replace = T))
      new_data=rbind(new_data,data[id,,drop=F])
      cancer_sampling=c(as.character(cancer_sampling),as.character(cancer[id]))
    }
  }
  data=new_data
  cancer=as.factor(cancer_sampling)
  return(list(data,cancer))
}

# Plots  ------------------------------------------------------------------

# function to plot the confusion matrix
# Args: 
#   confusion_matrix: confusion matrix to visualize
#   labels: vector containing the predicted labels for each sample
# Optional args: 
#   low_grad: color corresponding to the lowest values in the confusion matrix plot; high_grad: color corresponding to the highest values in the confusion matrix plot;
#   title: title of the confusion matrix plot; family_type: font used in the confusion matrix plot; classification: indicates the tumor types included in the analysis
plot_confusion_matrix_ggplot=function(confusion_matrix,labels,low_grad="#ece2f0",high_grad="#1c9099",title="",family_type="Times",classification="LCNECvsACvsTC"){
  x <- melt(confusion_matrix)
  names(x) <- c("iplan","vplan","value")
  n_intances=summary(labels)
  
  #get proportions
  x$value2<-rep(0,nrow(x))
  x$value2<- sapply(1:nrow(x), function(i) round(x$value[i]/n_intances[as.character(x$iplan[i])],2))
  
  # Colors and labels specifications
  if(classification=="LCNECvsACvsTC"){
    x$iplan <- factor(x$iplan, levels=c("LCNEC","Atypical","Typical"), labels=c("LCNEC","Atypical","Typical"))
    x$vplan <- factor(x$vplan, levels=c("Typical","Atypical","LCNEC","UC"), labels=c("Typical","Atypical","LCNEC","Unclassified"))
    x_col=c("black","#ff0000ff","#4abaffff",rgb(0.5,0.5,0.5))
    y_col=c("#4abaffff","#ff0000ff","black")
  }else if(classification=="LCNECvsPCA"){
    x$iplan <- factor(x$iplan, levels=c("LCNEC","PCA"), labels=c("LCNEC","PCA"))
    x$vplan <- factor(x$vplan, levels=c("PCA","LCNEC","UC"), labels=c("PCA","LCNEC","Unclassified"))
    x_col=c("black","#4abaffff",rgb(0.5,0.5,0.5))
    y_col=c("#4abaffff","black")
  }else if(classification=="LCNECvsAC"){
    x$iplan <- factor(x$iplan, levels=c("LCNEC","Atypical"), labels=c("LCNEC","Atypical"))
    x$vplan <- factor(x$vplan, levels=c("Atypical","LCNEC","UC"), labels=c("Atypical","LCNEC","Unclassified"))
    x_col=c("#ff0000ff","#4abaffff",rgb(0.5,0.5,0.5))
    y_col=c("#4abaffff","#ff0000ff")
  }else if(classification=="LCNECvsTC"){
    x$iplan <- factor(x$iplan, levels=c("LCNEC","Typical"), labels=c("LCNEC","Typical"))
    x$vplan <- factor(x$vplan, levels=c("Typical","LCNEC","UC"), labels=c("Typical","LCNEC","Unclassified"))
    x_col=c("#000000a0","#4abaffff",rgb(0.5,0.5,0.5))
    y_col=c("#4abaffff","#000000a0")
  }else if(classification=="ACvsTC"){
    x$iplan <- factor(x$iplan, levels=c("Atypical","Typical"), labels=c("Atypical","Typical"))
    x$vplan <- factor(x$vplan, levels=c("Typical","Atypical","UC"), labels=c("Typical","Atypical","Unclassified"))
    x_col=c("black","#ff0000ff",rgb(0.5,0.5,0.5))
    y_col=c("#ff0000ff","black")
  }
  
  plot=ggplot(x, aes(vplan, iplan),cex=0.1) +
    geom_point(aes(size = value2,colour=value2), alpha=1) +
    geom_text(aes(label = value2),size=4, color="black",family=family_type) +
    scale_size(range = c(4,18),guide = 'none') +
    scale_colour_gradient("Legend", low = low_grad, high = high_grad) + 
    labs(x="", y="")+theme_classic()+ggtitle(title)+
    theme(axis.text.x = element_text(angle = 20,hjust = 1,family=family_type,size = 12,color = x_col), 
          axis.text.y = element_text(angle = 0, hjust = 0.5,family=family_type,size = 12,color = y_col),
          legend.title = element_text(family=family_type), 
          title =  element_text(family=family_type), 
          text = element_text(family=family_type)) 
  
  print(plot)
  
}




